/*
  Copyright: G. Lipari
  e-mail:    g.lipari@sssup.it

  Project 1: Sudoku

  1) Modificare la funzione main() in modo che sia possibile 
  eseguire il programma con le seguenti opzioni
  
  - batch
  
  $ ./sudoku <filename> 

  esegue e tenta di risolvere tutti i sudoku, contando quanti sono
  effettivamente quelli risolti sul totale.

  - interactive: 
  
  $ ./sudoku <filename> -i <record_number> 

  esegue il programma il maniera interattiva, caricando il sudoku alla
  posizione 6 nel file di input, e fermandosi dopo ogni regola (come
  adesso)

  2) Aggiungere una o più regole per risolvere tutti i sudoku del file

  3) Aggiungere la modalità "creazione"

    $ ./sudoku -c <seed> <num_elem>

  crea un sudoku risolvibile utilizzando il seme random specificato in
  <seed> e contenente <num_elem> caselle non vuote. Verificare che il
  sudoku generato sia risolvibile. 

  Opzionale: creare un sudoku simmetrico (ad esempio per riflessione o
  per rotazione).

  Il sudoku generato viene stampato sullo standard output.  
*/



#include <stdio.h>
#include <stdlib.h>
///
/// Data structure to hold all variables for solving a sudoku
///
typedef struct __sudoku_type {
    ///
    /// global table
    /// 
    int table[9][9];

    /// 
    /// possible values for each cell
    /// feasible[i][j][k] = 1    if   k+1 is a feasible value for cell (i, j)
    /// feasible[i][j][k] = 0    if   k+1 is not a feasible value for cell (i, j)
    ///  
    int feasible[9][9][9];

    ///
    /// number of remaining empty cells
    ///
    int num_empty;

    ///
    /// used to see if something has changed by applying one rule
    ///
    int change;
} sudoku_t;


void select_feasible(sudoku_t *s);

///
/// helper function, returns a number c in [0, 8] that is the number of
/// 3x3 square that includes the [i, j] cell, and k in [0, 8] that is the 
/// cell number inside the square
/// 
void get_square(int i, int j, int *c, int *k)
{
    *c = (i/3)*3 + (j/3);
    *k = i%3*3 + j%3;
}


///
/// helper function gets the k-th cell of square c, and return 
/// coordinates in [i, j]
///
void get_cell(int c, int k, int *i, int *j)
{
    *i = (c / 3) * 3 + k / 3;
    *j = (c % 3) * 3 + k % 3;  
}

///
/// read table number n from file  
///
void read_table(sudoku_t *s, char *file, int n) 
{
    int i, j, k;
    char buff[100];

    FILE *f = fopen(file, "r");

    if (!f) {
        perror("Cannot open file");
        exit(-1);
    }

    for (i=0; i<n-1; i++) {
        for (j=0; j<10; j++) {
            fgets(buff, 100, f);
            if (feof(f)) {
                printf("Less than %i records in %s\n", n, file);
                exit(-1);
            }
        }
    }
    // initialize table
    s->num_empty = 0;
    s->change = 0;

    // read first row (not useful)
    fgets(buff, 100, f);
    for (i=0; i<9; i++) {
        fgets(buff, 100, f);

        for (j=0; j<9; j++) {
            s->table[i][j] = buff[j] - '0';

            if (s->table[i][j] == 0) s->num_empty++;
            if (feof(f)) {
                printf("records %i is incomplete\n", n);
                exit(-1);
            }
        }
    }

    /// initialize all possible values to 1
    for (i=0; i<9; i++) 
        for (j=0; j<9; j++)
            for (k=0; k<9; k++)
                s->feasible[i][j][k] = 1;
}

///
/// Print the diagram
///
void pretty_print_table(sudoku_t *s)
{
    int i, j;

    for (i=0; i<9; i++) {
        if (i%3 == 0) printf("+---+---+---+\n");
        for (j=0; j<9; j++) {
            if (j%3 == 0) printf("|");
            if (s->table[i][j] == 0) printf(" "); 
            else printf("%d", s->table[i][j]);
        }
        printf("|\n");
    }
    printf("+---+---+---+\n");
    printf("Empty cells: %d\n", s->num_empty);

}

// prototype
int count_feasible(sudoku_t *s, int i, int j, int array[]);

///
/// Print debugging information
///
void debug_print_feasible(sudoku_t *s)
{
    int i, j, k;
    int count;
    int array[9];
    int c, h;
    int i2, j2;

    for (i=0; i<9; i++) {
        for (j=0; j<9; j++) {
            if (s->table[i][j] == 0) {
                count = count_feasible(s, i, j, array);
                printf("Cell (%d, %d):\n", i+1, j+1);
                // get_square(i, j, &c, &h);
                // get_cell(c, h, &i2, &j2);
                // printf("     square = %d, h  = %d\n", c, h);
                // printf("     i2     = %d, j2 = %d\n", i2, j2);
                printf("     feasible = %d  {", count);
                for (k=0; k<count; k++) printf("%d, ", array[k]);
                printf("}\n");
            }
        }
    }
}


///
/// Count how many possibilities for every cell
/// returns an array with the possibilities
///
int count_feasible(sudoku_t *s, int i, int j, int array[])
{
    int k;
    int count = 0;
    int h = 0;

    for (k=0; k<9; k++) {
        count += s->feasible[i][j][k];
        if (s->feasible[i][j][k] == 1) array[h++] = k+1;
    }
    return count;
}


///
/// Elementary rule: if there is only one possibility for a cell,
/// set the value in the table, and return
/// 
void set_unique(sudoku_t *s)
{
    int i, j, k;
    int count;
    int array[9];

    for (i=0; i<9; i++) {
        for (j=0; j<9; j++) {
            if (s->table[i][j] == 0) {
                count = count_feasible(s, i, j, array);
	
                if (count == 1) {
                    s->table[i][j] = array[0];
                    s->change = 1;
                    s->num_empty--;
                    select_feasible(s);
                }
            }
        }
    }
}

// Second rule.  Look at every row (and column and 3x3 square), 
// to see if a certain number can only go in one place 
void look_at_number(sudoku_t *s, int n)
{
    int i, j, k, c, h;
    // this counts in how many files the number under test can go.
    // if it is 1, then it can only go in one place, so set it!
    int count;

    for (i = 0; i<9; i++) {
	count = 0;
    	for (j=0; j<9; j++) {
            if (s->table[i][j] == n) {
                count = -1;
                break;
            }
            else if (s->table[i][j] == 0) {
                count += s->feasible[i][j][n-1];
                if (s->feasible[i][j][n-1] == 1) h = j;
            }
    	}
    	if (count == 1) {
            s->table[i][h] = n;
            s->num_empty--;
            s->change = 1;
            select_feasible(s);
    	}
    }
  
    for (j=0; j<9; j++) {
    	count = 0;
    	for (i=0; i<9; i++) {
            if (s->table[i][j] == n) {
                count = -1;
                break;
            }
            else if (s->table[i][j] == 0) {
                count += s->feasible[i][j][n-1];
                if (s->feasible[i][j][n-1] == 1) h = i;
            }
	}
	if (count == 1) {
            s->table[h][j] = n;
            s->num_empty--;
            s->change = 1;
            select_feasible(s);
	}
    }

    for (c=0; c<9; c++) {
	count = 0;
	for (k=0; k<9; k++) {
            get_cell(c, k, &i, &j);
            if (s->table[i][j] == n) {
                count = -1;
                break;
            }
            else if (s->table[i][j] == 0) {
                count += s->feasible[i][j][n-1];
                if (s->feasible[i][j][n-1] == 1) h = k;
            }
	}
	if (count == 1) {
            get_cell(c, h, &i, &j);
            s->table[i][j] = n;
            s->num_empty--;
            s->change = 1;
            select_feasible(s);
	}
    }
}

///
/// Eliminate numbers existing on the same row, on the same 
/// column and on the same square from the feasible array
///
void cancel_simple(sudoku_t *s, int i, int j)
{
    int k, num;
    int c;
    int i2, j2, k2;

    if (s->table[i][j] != 0) return;

    get_square(i, j, &c, &k2);  // c is the square index of cell (i, j)

    for (k=0; k<9; k++) {
        /// eliminate existing elements on the same row
        num = s->table[i][k];
        if (num != 0) 
            s->feasible[i][j][num-1] = 0;

        /// eliminate existing elements on the same column
        num = s->table[k][j];
        if (num != 0)
            s->feasible[i][j][num-1] = 0;
    
        /// eliminate existing elements on the same square
        get_cell(c, k, &i2, &j2);
        num = s->table[i2][j2];
        if (num != 0)
            s->feasible[i][j][num-1] = 0;   
    }
}

///
/// Apply the elimination rules (could be more than one...)
/// 
void select_feasible(sudoku_t *s)
{
    int i; int j;

    for (i=0; i<9; i++) 
        for (j=0; j<9; j++) {
            cancel_simple(s, i, j);
        }
}

///
/// The main function
/// 
int main(int argc, char *argv[])
{
    sudoku_t sudoku;
    int count = 0;
    int n;

    if (argc < 3) {
        printf("Usage: %s <filename> <record number>\n", argv[0]);
        exit(0);
    }

    read_table(&sudoku, argv[1], atoi(argv[2]));

    while (sudoku.num_empty > 0) {
        pretty_print_table(&sudoku);

        printf("Press <enter> to continue\n");
        getchar();

        printf("STEP %d: select uniques\n", count++);
        select_feasible(&sudoku);
        //debug_print_feasible(&sudoku);
        set_unique(&sudoku);

        pretty_print_table(&sudoku);
        printf("Press <enter> to continue\n");
        getchar();

        printf("STEP %d: look for numbers\n", count++);
        select_feasible(&sudoku);
        for (n=1; n<10; n++) look_at_number(&sudoku, n);

        if (sudoku.change == 0) break;
        else 
            sudoku.change = 0;
    }

    if (sudoku.num_empty > 0) printf("Could not find any other possibility!\n");
    else printf("Completed!\n");
    pretty_print_table(&sudoku);

    return 0;
}
